#!/usr/bin/env python3

import argparse
import json
import os
import shutil
from tempfile import NamedTemporaryFile

from rich import print as rprint
from rich import print_json
from rich.markup import escape

class Concordance(object):
    def __init__(self, pile):
        self.assertions = []
        for fh in pile:
            c = json.load(fh)
            self.assertions.extend(c['assertions'])

    def __iter__(self):
        return iter(self.assertions)

    def getTarget(self, source):
        for a in self.assertions:
            if a["source"] == source:
                return a["target"]

    def getInfo(self, source):
        for a in self.assertions:
            if a["source"] == source:
                return a


TABLEAU_TEMPLATE = """
IF NOT ISNULL([[red]{source}[/red]]) THEN
    [red]{source}[/red]
ELSE
    [green]{target}[/green]
END
""".strip()


def main():

    methods = {
        "consistency": consistency,
        "convert": convert_instance,
        "augment": augment_instance,
        "search": search,
        "similar": similar,
        "dump": dump_mappable,
        "tableau": convert_tableau,
    }

    parser = argparse.ArgumentParser(description="XBRL concordance processor")
    parser.add_argument("MODE", choices=sorted(methods.keys()), help="Operation to carry out")
    parser.add_argument("--concordance", help="Path to concordance file", type=argparse.FileType('r', encoding='UTF-8'), nargs="+")
    parser.add_argument("--data-source", help="Path to useful data", type=argparse.FileType('r', encoding='UTF-8'))
    parser.add_argument("--start-with", help="Path to input instance / inline XBRL file", type=argparse.FileType('r', encoding='UTF-8'))
    parser.add_argument("--save-to", help="Path to output instance / inline XBRL file", type=argparse.FileType('w', encoding='UTF-8'))
    parser.add_argument("-s", help="Search text")
    cli = parser.parse_args()

    command = methods[cli.MODE]
    command(cli)


def consistency(cli):
    if len(cli.concordance) != 2:
        print("Must specify two concordances to check consistency")
        print(cli.concordance)
        raise SystemExit(1)

    print(f"Loading {cli.concordance[0].name}")
    c1 = json.load(cli.concordance[0])
    print(f"Loading {cli.concordance[1].name}")
    c2 = json.load(cli.concordance[1])
    print()

    a1list = c1["assertions"]
    a2list = c2["assertions"]

    for i, alist in enumerate((a1list, a2list), start = 1):
        rprint(f"Concordance {i} contains [blue]{len(alist)}[/blue] assertions.")
    print()

    consistent = []
    inconsistent = []
    complimentary = []
    b_used = []
    for a in c1["assertions"]:
        used = False
        for b in c2["assertions"]:
            if a == b:
                consistent.append(a)
                used = True
                b_used.append(b)
            elif a["source"] == b["source"] and a["target"] != b["target"]:
                inconsistent.append((a,b))
                used = True
                b_used.append(b)
        if not used:
            complimentary.append(a)
    
    complimentary.extend([b for b in c2["assertions"] if b not in b_used])

    vword = "[green]PASS" if len(inconsistent) == 0 else "[red]FAIL"
    rprint(f"Overall result: {vword}")
    rprint(f" - [blue]{len(consistent)}[/blue] consistent duplicate mapping found")
    rprint(f" - [blue]{len(inconsistent)}[/blue] inconsistent duplicate mappings found")
    rprint(f" - [blue]{len(complimentary)}[/blue] complimentary mappings found")
    print()
    print("Inconsistent concordance assertions: ")
    for i, (a,b) in enumerate(inconsistent, start=1):
        print(f"Pair: {i}")
        print_json(data=a)
        print_json(data=b)


def rewrite_file(file_path, target_path, *rewrite_pairs):
    with open(file_path) as current, NamedTemporaryFile(
        mode="w",
        dir=os.path.dirname(file_path),
        prefix=os.path.basename(file_path),
        delete=False,
    ) as working:
        for line in current:
            new_line = line
            for pair in rewrite_pairs:
                new_line = string_switch(new_line, pair)
            working.write(new_line)
    shutil.move(working.name, target_path)


def string_switch(old_string, pair):
    return old_string.replace(pair[0], pair[1])


def augment_instance(cli):
    C = Concordance(cli.concordance)
    d = json.load(cli.data_source)

    augment_pairs = []
    for id, stuff in d["facts"].items():
        concept = stuff["dimensions"]["concept"]
        value = stuff["value"]
        augment_pairs.append((f"<!-- @{C.getTarget(concept)}@ -->", value))

    for line in cli.start_with:
        new_line = line
        for a in augment_pairs:
            new_line = string_switch(new_line, a)
        cli.save_to.write(new_line)

def convert_instance(cli):
    C = Concordance(cli.concordance)
    print(f"Loading OIM data from {cli.start_with.name}")
    oim = json.load(cli.start_with)

    print("Using concordance to update OIM data")
    for stuff in oim["facts"].values():
        concept = stuff["dimensions"]["concept"]
        replacement = C.getTarget(concept)
        if replacement is not None:
            stuff["dimensions"]["concept"] = replacement
        else:
            print(f"No replacement found for {concept}")

    print(f"Writing out updated OIM data to {cli.save_to.name}")
    json.dump(oim, cli.save_to, indent = 1)

def dump_mappable(cli):
    C = Concordance(cli.concordance)
    d = json.load(cli.data_source)

    for fact_id, stuff in d["facts"].items():
        concept = stuff["dimensions"]["concept"]
        value = stuff["value"]
        other = C.getTarget(concept)
        if other is not None:
            rprint(f"Original with value {value} and concept {concept} ({fact_id})")
            rprint(f"Concordant fact {fact_id} with value {value} and concept {other}")
            print_json(data=C.getInfo(concept))
        else:
            rprint(f"Fact {fact_id} has no applicable concordance assertion.")


def convert_tableau(cli):
    C = Concordance(cli.concordance)
    for i, a in enumerate(C, start=1):
        if isinstance(a["source"], str) and "rules" not in a:
            print(f"Assertion {i} translated to a Tableau formula:")
            rprint(TABLEAU_TEMPLATE.format(source = escape(a["source"]), target = escape(a["target"])))
        else:
            print(f"Assertion {i} is complex and cannot be translated")
            print_json(data=a)
        print()


def search(cli):
    C = Concordance(cli.concordance)
    for a in C:   
        if isinstance(a["source"], str) and cli.s.lower() in a["source"].lower():
            rprint("Text match: [blue]" + escape(a["target"]) + "[/blue]")

def similar(cli):
    C = Concordance(cli.concordance)
    for a in C:
        if not "type" in a:
            continue
        if a["type"] == "similar" and cli.s == a["source"]:
            rprint("Similar: [blue]" + escape(a["target"]) + "[/blue]")

if __name__ == '__main__':
    main()